# _bayes.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>


import torch

from ._utils import _cast_as_tensor
from ._utils import _update_parameter
from ._utils import _check_parameter
from ._utils import _reshape_weights

from .distributions._distribution import Distribution


class BayesMixin(torch.nn.Module):
	def _reset_cache(self):
		"""Reset the internally stored statistics.

		This method is meant to only be called internally. It resets the
		stored statistics used to update the model parameters as well as
		recalculates the cached values meant to speed up log probability
		calculations.
		"""

		if self._initialized == False:
			return

		self.register_buffer("_w_sum", torch.zeros(self.k, device=self.device))
		self.register_buffer("_log_priors", torch.log(self.priors))

	def _emission_matrix(self, X, priors=None):
		"""Return the emission/responsibility matrix.

		This method returns the log probability of each example under each
		distribution contained in the model with the log prior probability
		of each component added.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate. 

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, self.k)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities). This can be used to assign labels to observatons
			by setting one of the probabilities for an observation to 1.0.
			Note that this can be used to assign hard labels, but does not
			have the same semantics for soft labels, in that it only
			influences the initial estimate of an observation being generated
			by a component, not gives a target. Default is None.

	
		Returns
		-------
		e: torch.Tensor, shape=(-1, self.k)
			A set of log probabilities for each example under each distribution.
		"""

		X = _check_parameter(_cast_as_tensor(X), "X", ndim=2, 
			shape=(-1, self.d), check_parameter=self.check_data)

		priors = _check_parameter(_cast_as_tensor(priors), "priors",
			ndim=2, shape=(X.shape[0], self.k), min_value=0.0, max_value=1.0,
			value_sum=1.0, value_sum_dim=-1, check_parameter=self.check_data)

		d = X.shape[0]
		e = torch.empty(d, self.k, device=self.device, dtype=self.dtype)
		for i, d in enumerate(self.distributions):
			e[:, i] = d.log_probability(X)

		if priors is not None:
			e += torch.log(priors)

		return e + self._log_priors

	def probability(self, X, priors=None):
		"""Calculate the probability of each example.

		This method calculates the probability of each example given the
		parameters of the distribution. The examples must be given in a 2D
		format.

		Note: This differs from some other probability calculation
		functions, like those in torch.distributions, because it is not
		returning the probability of each feature independently, but rather
		the total probability of the entire example.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, self.k)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities). This can be used to assign labels to observatons
			by setting one of the probabilities for an observation to 1.0.
			Note that this can be used to assign hard labels, but does not
			have the same semantics for soft labels, in that it only
			influences the initial estimate of an observation being generated
			by a component, not gives a target. Default is None.


		Returns
		-------
		prob: torch.Tensor, shape=(-1,)
			The probability of each example.
		"""

		return torch.exp(self.log_probability(X, priors=priors))

	def log_probability(self, X, priors=None):
		"""Calculate the log probability of each example.

		This method calculates the log probability of each example given the
		parameters of the distribution. The examples must be given in a 2D
		format. For a Bernoulli distribution, each entry in the data must
		be either 0 or 1.

		Note: This differs from some other log probability calculation
		functions, like those in torch.distributions, because it is not
		returning the log probability of each feature independently, but rather
		the total log probability of the entire example.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, self.k)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities). This can be used to assign labels to observatons
			by setting one of the probabilities for an observation to 1.0.
			Note that this can be used to assign hard labels, but does not
			have the same semantics for soft labels, in that it only
			influences the initial estimate of an observation being generated
			by a component, not gives a target. Default is None.


		Returns
		-------
		logp: torch.Tensor, shape=(-1,)
			The log probability of each example.
		"""

		e = self._emission_matrix(X, priors=priors)
		return torch.logsumexp(e, dim=1)

	def predict(self, X, priors=None):
		"""Calculate the label assignment for each example.

		This method calculates the label for each example as the most likely
		component after factoring in the prior probability.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to summarize.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, self.k)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities). This can be used to assign labels to observatons
			by setting one of the probabilities for an observation to 1.0.
			Note that this can be used to assign hard labels, but does not
			have the same semantics for soft labels, in that it only
			influences the initial estimate of an observation being generated
			by a component, not gives a target. Default is None.


		Returns
		-------
		y: torch.Tensor, shape=(-1,)
			The predicted label for each example.
		"""

		e = self._emission_matrix(X, priors=priors)
		return torch.argmax(e, dim=1)

	def predict_proba(self, X, priors=None):
		"""Calculate the posterior probabilities for each example.

		This method calculates the posterior probabilities for each example
		under each component of the model after factoring in the prior 
		probability and normalizing across all the components.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to summarize.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, self.k)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities). This can be used to assign labels to observatons
			by setting one of the probabilities for an observation to 1.0.
			Note that this can be used to assign hard labels, but does not
			have the same semantics for soft labels, in that it only
			influences the initial estimate of an observation being generated
			by a component, not gives a target. Default is None.


		Returns
		-------
		y: torch.Tensor, shape=(-1, self.k)
			The posterior probabilities for each example under each component.
		"""

		e = self._emission_matrix(X, priors=priors)
		return torch.exp(e - torch.logsumexp(e, dim=1, keepdims=True))
		
	def predict_log_proba(self, X, priors=None):
		"""Calculate the log posterior probabilities for each example.

		This method calculates the log posterior probabilities for each example
		under each component of the model after factoring in the prior 
		probability and normalizing across all the components.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to summarize.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, self.k)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities). This can be used to assign labels to observatons
			by setting one of the probabilities for an observation to 1.0.
			Note that this can be used to assign hard labels, but does not
			have the same semantics for soft labels, in that it only
			influences the initial estimate of an observation being generated
			by a component, not gives a target. Default is None.


		Returns
		-------
		y: torch.Tensor, shape=(-1, self.k)
			The log posterior probabilities for each example under each 
			component.
		"""

		e = self._emission_matrix(X, priors=priors) 
		return e - torch.logsumexp(e, dim=1, keepdims=True)

	def from_summaries(self):
		"""Update the model parameters given the extracted statistics.

		This method uses calculated statistics from calls to the `summarize`
		method to update the distribution parameters. Hyperparameters for the
		update are passed in at initialization time.

		Note: Internally, a call to `fit` is just a successive call to the
		`summarize` method followed by the `from_summaries` method.
		"""

		for d in self.distributions:
			d.from_summaries()

		if self.frozen == True:
			return

		priors = self._w_sum / torch.sum(self._w_sum)

		_update_parameter(self.priors, priors, self.inertia)
		self._reset_cache()



# markov_chain.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

import torch

from ._utils import _cast_as_tensor
from ._utils import _update_parameter
from ._utils import _check_parameter
from ._utils import _reshape_weights

from .distributions._distribution import Distribution
from .distributions import Categorical
from .distributions import ConditionalCategorical


class MarkovChain(Distribution):
	"""A Markov chain.

	A Markov chain is the simplest sequential model which factorizes the
	joint probability distribution P(X_{0} ... X_{t}) along a chain into the
	product of a marginal distribution P(X_{0}) P(X_{1} | X_{0}) ... with
	k conditional probability distributions for a k-th order Markov chain.

	Despite sometimes being thought of as an independent model, Markov chains
	are probability distributions over sequences just like hidden Markov
	models. Because a Markov chain has the same theoretical properties as a
	probability distribution, it can be used in any situation that a simpler 
	distribution could, such as an emission distribution for a HMM or a 
	component of a Bayes classifier.


	Parameters
	----------
	distributions: tuple or list or None
		A set of distribution objects. These objects do not need to be
		initialized, i.e., can be "Categorical()". 

	k: int or None
		The number of conditional distributions to include in the chain, also
		the number of steps back to model in the sequence. This must be passed
		in if the distributions are not passed in.

	n_categories: list, tuple, or None
		A list or tuple containing the number of categories that each feature
		has. 

	inertia: float, [0, 1], optional
		Indicates the proportion of the update to apply to the parameters
		during training. When the inertia is 0.0, the update is applied in
		its entirety and the previous parameters are ignored. When the
		inertia is 1.0, the update is entirely ignored and the previous
		parameters are kept, equivalently to if the parameters were frozen.

	frozen: bool, optional
		Whether all the parameters associated with this distribution are frozen.
		If you want to freeze individual pameters, or individual values in those
		parameters, you must modify the `frozen` attribute of the tensor or
		parameter directly. Default is False.

	check_data: bool, optional
		Whether to check properties of the data and potentially recast it to
		torch.tensors. This does not prevent checking of parameters but can
		slightly speed up computation when you know that your inputs are valid.
		Setting this to False is also necessary for compiling.
	"""

	def __init__(self, distributions=None, k=None, n_categories=None, 
		inertia=0.0, frozen=False, check_data=True):
		super().__init__(inertia=inertia, frozen=frozen, check_data=check_data)
		self.name = "MarkovChain"

		self.distributions = _check_parameter(distributions, "distributions",
			dtypes=(list, tuple))
		self.k = _check_parameter(_cast_as_tensor(k, dtype=torch.int32), "k",
			ndim=0)
		self.n_categories = _check_parameter(n_categories, "n_categories",
			dtypes=(list, tuple))

		if distributions is None and k is None:
			raise ValueError("Must provide one of 'distributions', or 'k'.")

		if distributions is not None:
			self.k = len(distributions) - 1

		self.d = None
		self._initialized = distributions is not None and distributions[0]._initialized
		self._reset_cache()

	def _initialize(self, d, n_categories):
		"""Initialize the probability distribution.

		This method is meant to only be called internally. It initializes the
		parameters of the distribution and stores its dimensionality. For more
		complex methods, this function will do more.


		Parameters
		----------
		d: int
			The dimensionality the distribution is being initialized to.

		n_categories: int
			The maximum number of categories to model. This single number is
			used as the maximum across all features and all timesteps.
		"""

		if self.distributions is None:
			self.distributions = [Categorical()]
			self.distributions[0]._initialize(d, max(n_categories))

			for i in range(self.k):
				distribution = ConditionalCategorical()
				distribution._initialize(d, [[n_categories[j]]*(i+2) 
					for j in range(d)])

				self.distributions.append(distribution)

		self.n_categories = n_categories
		self._initialized = True
		super()._initialize(d)

	def _reset_cache(self):
		"""Reset the internally stored statistics.

		This method is meant to only be called internally. It resets the
		stored statistics used to update the model parameters as well as
		recalculates the cached values meant to speed up log probability
		calculations.
		"""

		if self._initialized:
			for distribution in self.distributions:
				distribution._reset_cache()

	def sample(self, n):
		"""Sample from the probability distribution.

		This method will return `n` samples generated from the underlying
		probability distribution. For a mixture model, this involves first
		sampling the component using the prior probabilities, and then sampling
		from the chosen distribution.


		Parameters
		----------
		n: int
			The number of samples to generate.
		

		Returns
		-------
		X: torch.tensor, shape=(n, self.d)
			Randomly generated samples.
		"""

		X = [self.distributions[0].sample(n)]

		for distribution in self.distributions[1:]:
			X_ = torch.stack(X).permute(1, 0, 2)
			samples = distribution.sample(n, X_[:, -self.k-1:])
			X.append(samples)

		return torch.stack(X).permute(1, 0, 2)

	def log_probability(self, X):
		"""Calculate the log probability of each example.

		This method calculates the log probability of each example given the
		parameters of the distribution. The examples must be given in a 3D
		format.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, length, self.d)
			A set of examples to evaluate.

		Returns
		-------
		logp: torch.Tensor, shape=(-1,)
			The log probability of each example.
		"""


		X = _check_parameter(_cast_as_tensor(X), "X", ndim=3, 
			check_parameter=self.check_data)
		self.d = X.shape[1]

		logps = self.distributions[0].log_probability(X[:, 0])
		for i, distribution in enumerate(self.distributions[1:-1]):
			logps += distribution.log_probability(X[:, :i+2])

		for i in range(X.shape[1] - self.k):
			j = i + self.k + 1
			logps += self.distributions[-1].log_probability(X[:, i:j])

		return logps

	def fit(self, X, sample_weight=None):
		"""Fit the model to optionally weighted examples.

		This method will fit the provided distributions given the data and
		their weights. If only `k` has been provided, the relevant set of
		distributions will be initialized.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, length, self.d)
			A set of examples to evaluate. 

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.


		Returns
		-------
		self
		"""

		self.summarize(X, sample_weight=sample_weight)
		self.from_summaries()
		return self

	def summarize(self, X, sample_weight=None):
		"""Extract the sufficient statistics from a batch of data.

		This method calculates the sufficient statistics from optionally
		weighted data and adds them to the stored cache for each distribution
		in the network. Sample weights can either be provided as one
		value per example or as a 2D matrix of weights for each feature in
		each example.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, length, self.d)
			A set of examples to summarize.

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.


		Returns
		-------
		logp: torch.Tensor, shape=(-1,)
			The log probability of each example.
		"""

		if self.frozen:
			return

		X = _check_parameter(_cast_as_tensor(X), "X", ndim=3, 
			check_parameter=self.check_data)
		sample_weight = _check_parameter(_cast_as_tensor(sample_weight), 
			"sample_weight", min_value=0, ndim=(1, 2), 
			check_parameter=self.check_data)

		if not self._initialized:
			if self.n_categories is not None:
				n_keys = self.n_categories
			elif isinstance(X, torch.masked.MaskedTensor):
				n_keys = (torch.max(torch.max(X._masked_data, dim=0)[0], 
					dim=0)[0] + 1).type(torch.int32)
			else:
				n_keys = (torch.max(torch.max(X, dim=0)[0], dim=0)[0] + 1).type(
					torch.int32)

			self._initialize(len(X[0][0]), n_keys)

		if sample_weight is None:
			sample_weight = torch.ones_like(X[:, 0])
		elif len(sample_weight.shape) == 1: 
			sample_weight = sample_weight.reshape(-1, 1).expand(-1, X.shape[2])
		elif sample_weight.shape[1] == 1:
			sample_weight = sample_weight.expand(-1, X.shape[2])

		_check_parameter(_cast_as_tensor(sample_weight), "sample_weight", 
			min_value=0, ndim=2, shape=(X.shape[0], X.shape[2]), 
			check_parameter=self.check_data)

		self.distributions[0].summarize(X[:, 0], sample_weight=sample_weight)
		for i, distribution in enumerate(self.distributions[1:-1]):
			distribution.summarize(X[:, :i+2], sample_weight=sample_weight)

		distribution = self.distributions[-1]
		for i in range(X.shape[1] - self.k):
			j = i + self.k + 1
			distribution.summarize(X[:, i:j], sample_weight=sample_weight)

	def from_summaries(self):
		"""Update the model parameters given the extracted statistics.

		This method uses calculated statistics from calls to the `summarize`
		method to update the distribution parameters. Hyperparameters for the
		update are passed in at initialization time.

		Note: Internally, a call to `fit` is just a successive call to the
		`summarize` method followed by the `from_summaries` method.
		"""

		if self.frozen:
			return

		for distribution in self.distributions:
			distribution.from_summaries()



# test_kmeans.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import os
import numpy
import torch
import pytest

from pomegranate.kmeans import KMeans

from .distributions._utils import _test_initialization_raises_one_parameter
from .distributions._utils import _test_initialization
from .distributions._utils import _test_predictions
from .distributions._utils import _test_efd_from_summaries
from .distributions._utils import _test_raises

from .tools import assert_raises
from numpy.testing import assert_array_almost_equal


MIN_VALUE = None
MAX_VALUE = None
VALID_VALUE = 1.2
inf = float("inf")


@pytest.fixture
def X():
	return [[1, 2, 0],
	     [0, 0, 1],
	     [1, 1, 2],
	     [2, 2, 2],
	     [3, 1, 0],
	     [5, 1, 4],
	     [2, 1, 0],
	     [1, 0, 2],
	     [1, 1, 0],
	     [0, 2, 1],
	     [0, 0, 0]]


@pytest.fixture
def X_masked(X):
	mask = torch.tensor(numpy.array([
		[False, True,  True ],
		[True,  True,  False],
		[False, False, False],
		[True,  True,  True ],
		[False, True,  False],
		[True,  True,  True ],
		[False, False, False],
		[True,  False, True ],
		[True,  True,  True ],
		[True,  True,  True ],
		[True,  False, True ]]))

	X = torch.tensor(numpy.array(X))
	return torch.masked.MaskedTensor(X, mask=mask)


@pytest.fixture
def w():
	return [[1], [2], [0], [0], [5], [1], [2], [1], [1], [2], [0]]


@pytest.fixture
def model():
	centroids = [[1, 0, 1], [2, 1, -1]]
	return KMeans(centroids=centroids)


###


def test_initialization():
	model = KMeans(3)

	assert_raises(AttributeError, getattr, model, "_w_sum")
	assert_raises(AttributeError, getattr, model, "_xw_sum")


def test_initialization_raises():
	d = 2

	assert_raises(ValueError, KMeans)
	assert_raises(ValueError, KMeans, k=1)
	assert_raises(ValueError, KMeans, k=0)
	assert_raises(ValueError, KMeans, k=-1)
	assert_raises(ValueError, KMeans, k=2.4)

	assert_raises(ValueError, KMeans, d, inertia=-0.4)
	assert_raises(ValueError, KMeans, d, inertia=1.2)
	assert_raises(ValueError, KMeans, d, inertia=1.2, frozen="true")
	assert_raises(ValueError, KMeans, d, inertia=1.2, frozen=3)
	
	assert_raises(ValueError, KMeans, d, tol=-2)

	assert_raises(ValueError, KMeans, d, max_iter=0)
	assert_raises(ValueError, KMeans, d, max_iter=1.3)
	assert_raises(ValueError, KMeans, d, max_iter=-2)

	assert_raises(ValueError, KMeans, d, init="aaa")
	assert_raises(ValueError, KMeans, d, init=False)


def test_reset_cache(model, X):
	model.summarize(X)
	
	assert_array_almost_equal(model._w_sum, [[8., 8., 8.], [3., 3., 3.]])
	assert_array_almost_equal(model._xw_sum, [[10., 7., 12.], [6., 4., 0.,]])

	model._reset_cache()
	assert_array_almost_equal(model._w_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum, [[0., 0., 0.,], [0., 0., 0.]])	


def test_initialize(X):
	d = 4
	model = KMeans(d)

	assert model.d is None
	assert model._initialized == False
	assert_raises(AttributeError, getattr, model, "_w_sum")
	assert_raises(AttributeError, getattr, model, "_xw_sum")

	model._initialize(X)
	assert model._initialized == True
	assert model.centroids.shape == (4, 3)
	assert model.d == 3
	assert_array_almost_equal(model._w_sum,
		[[0., 0., 0.],
		 [0., 0., 0.],
		 [0., 0., 0.],
		 [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum,
		[[0., 0., 0.],
		 [0., 0., 0.],
		 [0., 0., 0.],
		 [0., 0., 0.]])



###


def test_predict(model, X):
	y_hat = model.predict(X)
	assert_array_almost_equal(y_hat, [1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0], 4)


def test_predict_raises(model, X):
	_test_raises(model, "predict", X)

	model = KMeans(2)
	_test_raises(model, "predict", X)


def test_distances(model, X):
	y_hat = model._distances(X)
	assert_array_almost_equal(y_hat,
		[[1.2910, 1.0000],
         [0.5774, 1.7321],
         [0.8165, 1.8257],
         [1.4142, 1.8257],
         [1.4142, 0.8165],
         [2.9439, 3.3665],
         [1.0000, 0.5774],
         [0.5774, 1.9149],
         [0.8165, 0.8165],
         [1.2910, 1.7321],
         [0.8165, 1.4142]], 4)


def test_distances_raises(model, X):
	_test_raises(model, "_distances", X)

	model = KMeans(2)
	_test_raises(model, "_distances", X)


###


def test_partial_summarize(model, X):
	model.summarize(X[:4])
	assert_array_almost_equal(model._w_sum, [[3., 3., 3.], [1., 1., 1.]])
	assert_array_almost_equal(model._xw_sum, [[3., 3., 5.], [1., 2., 0.]])

	model.summarize(X[4:])
	assert_array_almost_equal(model._w_sum, [[8., 8., 8.], [3., 3., 3.]])
	assert_array_almost_equal(model._xw_sum, [[10.,  7., 12.], [ 6.,  4.,  0.]])


def test_full_summarize(model, X):
	model.summarize(X)
	assert_array_almost_equal(model._w_sum, [[8., 8., 8.], [3., 3., 3.]])
	assert_array_almost_equal(model._xw_sum, [[10.,  7., 12.], [ 6.,  4.,  0.]])


def test_summarize_weighted(model, X, w):
	model.summarize(X, sample_weight=w)
	assert_array_almost_equal(model._w_sum, [[7., 7., 7.], [8., 8., 8.]])
	assert_array_almost_equal(model._xw_sum, [[ 7.,  6., 10.], [20.,  9.,  0.]])


def test_summarize_weighted_flat(model, X, w):
	w = numpy.array(w)[:,0] 

	model.summarize(X, sample_weight=w)
	assert_array_almost_equal(model._w_sum, [[7., 7., 7.], [8., 8., 8.]])
	assert_array_almost_equal(model._xw_sum, [[ 7.,  6., 10.], [20.,  9.,  0.]])


def test_summarize_weighted_2d(model, X):
	model.summarize(X, sample_weight=X)
	assert_array_almost_equal(model._w_sum, [[10.,  7., 12.], [6.,  4.,  0.]])
	assert_array_almost_equal(model._xw_sum, [[32., 11., 30.], [14.,  6.,  0.]])


def test_summarize_raises(model, X, w):
	assert_raises(ValueError, model.summarize, [X])
	assert_raises(ValueError, model.summarize, X[0])
	assert_raises((ValueError, TypeError), model.summarize, X[0][0])
	assert_raises(ValueError, model.summarize, [x[:-1] for x in X])

	assert_raises(ValueError, model.summarize, [X], w)
	assert_raises(ValueError, model.summarize, X, [w])
	assert_raises(ValueError, model.summarize, [X], [w])
	assert_raises(ValueError, model.summarize, X[:len(X)-1], w)
	assert_raises(ValueError, model.summarize, X, w[:len(w)-1])


def test_from_summaries(model, X):
	model.summarize(X)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model.centroids, 
		[[1.25    , 0.875   , 1.5     ],
         [2.      , 1.333333, 0.      ]])


def test_from_summaries_weighted(model, X, w):
	model.summarize(X, sample_weight=w)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model.centroids, 
		[[1.      , 0.857143, 1.428571],
         [2.5     , 1.125   , 0.      ]])


def test_from_summaries_null(model):
	centroids = torch.clone(model.centroids)

	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_raises(AssertionError, assert_array_almost_equal, model.centroids, centroids)


def test_from_summaries_inertia(X):
	model = KMeans(2, init='first-k', inertia=0.3)
	model.summarize(X)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model.centroids, 
		[[1.7  , 1.6  , 0.7  ],
         [0.35 , 0.175, 1.175]])


	centroids = X[:2]
	model = KMeans(centroids=centroids, inertia=1.0)
	model.summarize(X)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model.centroids, centroids)


def test_from_summaries_weighted_inertia(X, w):
	model = KMeans(2, init='first-k', inertia=0.3)
	model.summarize(X, sample_weight=w)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model.centroids, 
		[[1.816667, 1.475   , 0.35    ],
         [0.233333, 0.      , 1.233333]])


	centroids = X[:2]
	model = KMeans(centroids=centroids, inertia=1.0)
	model.summarize(X, sample_weight=w)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model.centroids, centroids)


def test_from_summaries_frozen(X):
	centroids = X[:2]
	model = KMeans(centroids=centroids, frozen=True)
	model.summarize(X, sample_weight=w)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model.centroids, centroids)


def test_fit(model, X):
	model.fit(X)

	assert_array_almost_equal(model._w_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model.centroids, 
		[[1.5, 1. , 2. ],
         [1.4, 1. , 0. ]])


def test_fit_weighted(model, X, w):
	model.fit(X, sample_weight=w)

	assert_array_almost_equal(model._w_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model.centroids, 
		[[0.428571, 1.      , 0.857143],
         [3.      , 1.      , 0.5     ]])


def test_fit_chain(X):
	model = KMeans(k=2, init='first-k').fit(X)

	assert_array_almost_equal(model._w_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model.centroids, 
		[[2.6     , 1.4     , 1.2     ],
         [0.5     , 0.666667, 1.      ]])


def test_fit_raises(model, X, w):
	assert_raises(ValueError, model.fit, [X])
	assert_raises(ValueError, model.fit, X[0])
	assert_raises((ValueError, TypeError), model.fit, X[0][0])
	assert_raises(ValueError, model.fit, [x[:-1] for x in X])

	assert_raises(ValueError, model.fit, [X], w)
	assert_raises(ValueError, model.fit, X, [w])
	assert_raises(ValueError, model.fit, [X], [w])
	assert_raises(ValueError, model.fit, X[:len(X)-1], w)
	assert_raises(ValueError, model.fit, X, w[:len(w)-1])


def test_serialization(X, model):
	torch.save(model, ".pytest.torch")
	model2 = torch.load(".pytest.torch")
	os.system("rm .pytest.torch")

	assert model is not model2

	assert_array_almost_equal(model2.centroids, model.centroids)
	assert_array_almost_equal(model2._w_sum, model._w_sum)
	assert_array_almost_equal(model2._xw_sum, model._xw_sum)
	assert_array_almost_equal(model2._centroid_sum, model._centroid_sum)

	assert_array_almost_equal(model2._distances(X), model._distances(X))


def test_masked_distances(model, X, X_masked):
	X = torch.tensor(numpy.array(X))
	mask = torch.ones_like(X).type(torch.bool)
	X_ = torch.masked.MaskedTensor(X, mask=mask)

	y_hat = model._distances(X_)
	assert_array_almost_equal(y_hat,
		[[1.2910, 1.0000],
         [0.5774, 1.7321],
         [0.8165, 1.8257],
         [1.4142, 1.8257],
         [1.4142, 0.8165],
         [2.9439, 3.3665],
         [1.0000, 0.5774],
         [0.5774, 1.9149],
         [0.8165, 0.8165],
         [1.2910, 1.7321],
         [0.8165, 1.4142]], 4)

	y_hat = model._distances(X_masked)
	assert_array_almost_equal(y_hat,
		[[1.7321, 1.7321],
         [1.0000, 1.7321],
         [   inf,    inf],
         [1.4142, 1.8257],
         [1.7321, 2.2361],
         [2.9439, 3.3665],
         [   inf,    inf],
         [0.7071, 2.3452],
         [0.8165, 0.8165],
         [1.2910, 1.7321],
         [1.0000, 1.7321]], 4)


def test_masked_summarize(model, X, X_masked, w):
	X = torch.tensor(numpy.array(X))
	mask = torch.ones_like(X).type(torch.bool)
	X_ = torch.masked.MaskedTensor(X, mask=mask)

	model.summarize(X_, sample_weight=w)
	assert_array_almost_equal(model._w_sum, [[7., 7., 7.], [8., 8., 8.]])
	assert_array_almost_equal(model._xw_sum, [[ 7.,  6., 10.], [20.,  9.,  0.]])

	model = KMeans(2)
	model.summarize(X_masked, sample_weight=w)
	assert_array_almost_equal(model._w_sum, [[ 4., 10.,  5.], [ 3.,  2.,  1.]])
	assert_array_almost_equal(model._xw_sum, [[ 6., 13.,  6.], [ 1.,  0.,  2.]])


def test_masked_from_summaries(model, X, X_masked, w):
	X = torch.tensor(numpy.array(X))
	mask = torch.ones_like(X).type(torch.bool)
	X_ = torch.masked.MaskedTensor(X, mask=mask)

	model.summarize(X_, sample_weight=w)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model.centroids, 
		[[1.      , 0.857143, 1.428571],
         [2.5     , 1.125   , 0.      ]])

	model = KMeans(centroids=[[1, 0, 1], [2, 1, 0]])
	model.summarize(X_masked, sample_weight=w)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model.centroids, 
		[[0.2     , 1.      , 1.333333],
         [3.      , 1.333333, 1.333333]])


def test_masked_fit(model, X, X_masked):
	X = torch.tensor(numpy.array(X))
	mask = torch.ones_like(X).type(torch.bool)
	X_ = torch.masked.MaskedTensor(X, mask=mask)

	model.fit(X_)

	assert_array_almost_equal(model._w_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model.centroids, 
		[[1.5, 1. , 2. ],
         [1.4, 1. , 0. ]])

	model = KMeans(centroids=[[1, 0, 1], [2, 1, 0]])
	model.fit(X_masked)

	assert_array_almost_equal(model._w_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model._xw_sum, [[0., 0., 0.], [0., 0., 0.]])
	assert_array_almost_equal(model.centroids, 
		[[0.4, 1.2, 0.6],
         [3.5, 1.5, 3. ]])


# BayesClassifier.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
import torch

from ._utils import _cast_as_tensor
from ._utils import _cast_as_parameter
from ._utils import _update_parameter
from ._utils import _check_parameter
from ._utils import _reshape_weights

from ._bayes import BayesMixin

from .distributions._distribution import Distribution


class BayesClassifier(BayesMixin, Distribution):
	"""A Bayes classifier object.

	A simple way to produce a classifier using probabilistic models is to plug
	them into Bayes' rule. Basically, inference is the same as the 'E' step in
	EM for mixture models. However, fitting can be significantly faster because
	instead of having to iteratively infer labels and learn parameters, you can
	just learn the parameters given the known labels. Because the learning step
	for most models are simple MLE estimates, this can be done extremely
	quickly.

	Although the most common distribution to use is a Gaussian with a diagonal
	covariance matrix, termed the Gaussian naive Bayes model, any probability
	distribution can be used. Here, you can just drop any distributions or
	probabilistic model in as long as it has the `log_probability`, `summarize`,
	and `from_samples` methods implemented.

	Further, the probabilistic models do not even need to be simple
	distributions. The distributions can be mixture models or hidden Markov
	models or Bayesian networks.


	Parameters
	----------
	distributions: tuple or list
		A set of distribution objects. These objects do not need to be
		initialized, i.e., can be "Normal()". 

	priors: tuple, numpy.ndarray, torch.Tensor, or None. shape=(k,), optional
		The prior probabilities over the given distributions. Default is None.

	inertia: float, [0, 1], optional
		Indicates the proportion of the update to apply to the parameters
		during training. When the inertia is 0.0, the update is applied in
		its entirety and the previous parameters are ignored. When the
		inertia is 1.0, the update is entirely ignored and the previous
		parameters are kept, equivalently to if the parameters were frozen.

	frozen: bool, optional
		Whether all the parameters associated with this distribution are frozen.
		If you want to freeze individual pameters, or individual values in those
		parameters, you must modify the `frozen` attribute of the tensor or
		parameter directly. Default is False.

	check_data: bool, optional
		Whether to check properties of the data and potentially recast it to
		torch.tensors. This does not prevent checking of parameters but can
		slightly speed up computation when you know that your inputs are valid.
		Setting this to False is also necessary for compiling. Default is True.
	"""

	def __init__(self, distributions, priors=None, inertia=0.0, frozen=False,
		check_data=True):
		super().__init__(inertia=inertia, frozen=frozen, check_data=check_data)
		self.name = "BayesClassifier"

		_check_parameter(distributions, "distributions", dtypes=(list, tuple, 
			numpy.array, torch.nn.ModuleList))
		self.distributions = torch.nn.ModuleList(distributions)

		self.priors = _check_parameter(_cast_as_parameter(priors), "priors", 
			min_value=0, max_value=1, ndim=1, value_sum=1.0, 
			shape=(len(distributions),))

		self.k = len(distributions)

		if all(d._initialized for d in distributions):
			self._initialized = True
			self.d = distributions[0].d
			if self.priors is None:
				self.priors = _cast_as_parameter(torch.ones(self.k) / self.k)

		else:
			self._initialized = False
			self.d = None
		
		self._reset_cache()

	def _initialize(self, d):
		"""Initialize the probability distribution.

		This method is meant to only be called internally. It initializes the
		parameters of the distribution and stores its dimensionality. For more
		complex methods, this function will do more.


		Parameters
		----------
		d: int
			The dimensionality the distribution is being initialized to.
		"""

		self.priors = _cast_as_parameter(torch.ones(self.k, dtype=self.dtype, 
			device=self.device) / self.k)

		self._initialized = True
		super()._initialize(d)

	def fit(self, X, y, sample_weight=None):
		"""Fit the model to optionally weighted examples.

		This method implements the core of the learning process. For a
		general Bayes model, this involves fitting each component of the model
		using the labels that are provided. 

		This method is largely a wrapper around the `summarize` and
		`from_summaries` methods. It's primary contribution is serving as a
		loop around these functions and to monitor convergence.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate. 

		y: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1,)
			A set of labels, one per example.

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.


		Returns
		-------
		self
		"""

		self.summarize(X, y, sample_weight=sample_weight)
		self.from_summaries()
		return self

	def summarize(self, X, y, sample_weight=None):
		"""Extract the sufficient statistics from a batch of data.

		This method calculates the sufficient statistics from optionally
		weighted data and adds them to the stored cache. The examples must be
		given in a 2D format. Sample weights can either be provided as one
		value per example or as a 2D matrix of weights for each feature in
		each example.

		For a Bayes' classifier, this step involves partitioning the data
		according to the labels and then training each component using MLE
		estimates.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to summarize.

		y: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1,)
			A set of labels, one per example.

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.
		"""

		X, sample_weight = super().summarize(X, sample_weight=sample_weight)
		y = _check_parameter(_cast_as_tensor(y), "y", min_value=0, 
			max_value=self.k-1, ndim=1, shape=(len(X),), 
			check_parameter=self.check_data)
		sample_weight = _check_parameter(sample_weight, "sample_weight", 
			min_value=0, shape=(-1, self.d), check_parameter=self.check_data)

		for j, d in enumerate(self.distributions):
			idx = y == j
			d.summarize(X[idx], sample_weight[idx])

			if self.frozen == False:
				self._w_sum[j] = self._w_sum[j] + sample_weight[idx].mean(
					dim=-1).sum()



# bayesian_network.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

import time
import numpy
import torch
import itertools
import networkx as nx

from ._utils import _cast_as_tensor
from ._utils import _cast_as_parameter
from ._utils import _update_parameter
from ._utils import _check_parameter
from ._utils import _reshape_weights

from .distributions._distribution import Distribution
from .distributions import Categorical
from .distributions import JointCategorical
from .distributions import ConditionalCategorical

from .factor_graph import FactorGraph


class BayesianNetwork(Distribution):
	"""A Bayesian network object.

	A Bayesian network is a probability distribution where dependencies between
	variables are explicitly encoded in a graph structure and the lack of an
	edge represents a conditional independence. These graphs are directed and
	typically must be acyclic, but this implementation allows for the networks
	to be cyclic as long as there is no assumption of convergence during
	inference.

	Inference is doing using loopy belief propagation along a factor graph
	representation. This is sometimes called the `sum-product` algorithm.
	It will yield exact results if the graph has a tree-like structure.
	Otherwise, if the graph is acyclic, it is guaranteed to converge but not
	necessarily to optimal results. If the graph is cyclic, there is no
	guarantee on convergence, but it is thought that the longer the loop is the 
	more likely one will get good results.

	Structure learning can be done using a variety of methods. 


	Parameters
	----------
	distributions: tuple or list or None
		A set of distribution objects. These do not need to be initialized,
		i.e. can be "Categorical()". Currently, they must be either Categorical
		or ConditionalCategorical distributions. If provided, they must be 
		consistent with the provided edges in that every conditional 
		distribution must have at least one parent in the provided structure. 
		Default is None.

	edges: tuple or list or None, optional
		A list or tuple of 2-tuples where the first element in the 2-tuple is 
		the parent distribution object and the second element is the child
		distribution object. If None, then no edges. Default is None.

	structure: tuple or list or None, optional
		A list or tuple of the parents for each distribution with a tuple
		containing no elements indicating a root node. For instance, 
		((), (0,), (), (0, 2)) would represent a graph with four nodes, 
		where the second distribution has the first distribution as a parent 
		and the fourth distribution has the first and third distributions as 
		parents. Use this only when you want new distribution objects to be
		created and fit when using the `fit` method. Default is None.

	max_iter: int, optional
		The number of iterations to do in the inference step. Default is 10.

	tol: float, optional
		The threshold at which to stop during fitting when the improvement
		goes under. Default is 1e-6.

	inertia: float, [0, 1], optional
		Indicates the proportion of the update to apply to the parameters
		during fitting. When the inertia is 0.0, the update is applied in
		its entirety and the previous parameters are ignored. When the
		inertia is 1.0, the update is entirely ignored and the previous
		parameters are kept, equivalently to if the parameters were frozen.

	frozen: bool, optional
		Whether all the parameters associated with this distribution are frozen.
		If you want to freeze individual pameters, or individual values in those
		parameters, you must modify the `frozen` attribute of the tensor or
		parameter directly. Default is False.

	check_data: bool, optional
		Whether to check properties of the data and potentially recast it to
		torch.tensors. This does not prevent checking of parameters but can
		slightly speed up computation when you know that your inputs are valid.
		Setting this to False is also necessary for compiling. Default is True.

	verbose: bool, optional
		Whether to print the improvement and timings during training.
	"""

	def __init__(self, distributions=None, edges=None, structure=None,
		algorithm=None, include_parents=None, exclude_parents=None, 
		max_parents=None, pseudocount=0.0, max_iter=20, tol=1e-6, inertia=0.0, 
		frozen=False, check_data=True, verbose=False):
		super().__init__(inertia=inertia, frozen=frozen, check_data=check_data)
		self.name = "BayesianNetwork"

		self.distributions = torch.nn.ModuleList([])
		self.edges = []
		self.structure = structure

		self._marginal_mapping = {}
		self._factor_mapping = {}
		self._distribution_mapping = {}
		self._parents = []
		self._factor_graph = FactorGraph(max_iter=max_iter, tol=tol, 
			frozen=frozen)

		self.algorithm = algorithm
		self.include_parents = include_parents
		self.exclude_parents = exclude_parents
		self.max_parents = max_parents
		
		self.pseudocount = pseudocount
		self.max_iter = max_iter
		self.tol = tol
		self.verbose = verbose

		self.d = 0
		self._initialized = (distributions is not None and 
			distributions[0]._initialized)
		self._reset_cache()


		if distributions is not None:
			_check_parameter(distributions, "factors", dtypes=(list, tuple))

			for distribution in distributions:
				self.add_distribution(distribution)

		if edges is not None:
			_check_parameter(edges, "edges", dtypes=(list, tuple))

			if isinstance(edges, (tuple, list)):
				for parent, child in edges:
					self.add_edge(parent, child)
			else:
				raise ValueError("Edges must be tuple or list.")

	def _initialize(self, d):
		self._initialized = True
		super()._initialize(d)

	def _reset_cache(self):
		return

	def add_distribution(self, distribution):
		"""Adds a distribution to the set of distributions.

		Adds a distribution to the set of distributions being stored in the
		BayesianNetwork object but also updates the underlying factor graph,
		adding in a marginal node and a factor node.

		
		Parameters
		----------
		distribution: pomegranate.distributions.Distribution
			A distribution object to include as a node. Currently must be a
			Categorical or a ConditionalCategorical distribution.
		"""

		if not isinstance(distribution, (Categorical, ConditionalCategorical)):
			raise ValueError("Must be Categorical or ConditionalCategorical")

		self.distributions.append(distribution)
		self.d += 1

		# Create the marginal distribution in the factor graph
		n_keys = distribution.probs[0].shape[-1]
		marginal = Categorical(torch.ones(1, n_keys) / n_keys)

		# Add the marginal and keep track of it
		self._factor_graph.add_marginal(marginal)
		self._marginal_mapping[distribution] = marginal
		self._parents.append(tuple())
		self._distribution_mapping[distribution] = len(self.distributions) - 1

		# If a conditional distribution, calculate the joint distribution
		# assuming uniform distributions for all the parents
		if isinstance(distribution, ConditionalCategorical):
			p = torch.clone(distribution.probs[0])
			p /= torch.prod(torch.tensor(p.shape[:-1]))
			factor = JointCategorical(p)
			
			# Add the factor to the factor graph
			self._factor_graph.add_factor(factor)
			self._factor_mapping[distribution] = factor
		else:
			# Otherwise, we add the univariate distribution as the factor
			self._factor_graph.add_factor(distribution)
			self._factor_mapping[distribution] = distribution
			factor = distribution

		self._factor_graph.add_edge(marginal, factor)


	def add_distributions(self, distributions):
		"""Adds several distributions to the set of distributions.

		
		Parameters
		----------
		distribution: iterable
			Any object that can be iterated over that returns distributions.
			Must be Categorical or ConditionalCategorical distributions.
		"""

		for distribution in distributions:
			self.add_distribution(distribution)

	def add_edge(self, parent, child):
		"""Adds a directed edge from the parent to the child node.

		Adds an edge to the list of edges associated with the BayesianNetwork
		object but also adds an appropriate edge in the underlying factor
		graph between the marginal node associated with the parent and
		the factor node associated with the child.


		Parameters
		----------
		parent: pomegranate.distributions.Distribution
			The distribution that the edge begins at.

		child: pomegranate.distributions.Distribution
			The distribution that the edge points to.
		"""

		if not isinstance(child, ConditionalCategorical):
			raise ValueError("Child distribution must be conditional.")

		if parent not in self._marginal_mapping:
			raise ValueError("Parent distribution must be in network.")

		if child not in self._marginal_mapping:
			raise ValueError("Child distribution must be in network.")

		if parent is child:
			raise ValueError("Cannot have self-loops.")

		self.edges.append((parent, child))

		p_idx = self._distribution_mapping[parent]
		c_idx = self._distribution_mapping[child]
		self._parents[c_idx] += (p_idx,)

		# Get the respective marginal and factor distributions and their idxs
		marginal = self._marginal_mapping[parent]
		factor = self._factor_mapping[child]

		m_idx = self._factor_graph._marginal_idxs[marginal]
		f_idx = self._factor_graph._factor_idxs[factor]

		# We need to keep this edge as the last one so pop it and re-add it
		m = self._factor_graph._factor_edges[f_idx].pop()
		f = self._factor_graph._marginal_edges[m_idx].pop()

		# Add the new edge and then the previous edge
		self._factor_graph.add_edge(marginal, factor)
		self._factor_graph._factor_edges[f_idx].append(m)
		self._factor_graph._marginal_edges[m_idx].append(f)

	def add_edges(self, edges):
		"""Adds several edges to the network at once.

		
		Parameters
		----------
		edges: iterable
			Any object that can be iterated over that returns tuples with
			a pair of distributions.
		"""

		for edge in edges:
			self.add_edge(*edge)

	def sample(self, n):
		"""Sample from the probability distribution.

		This method will return `n` samples generated from the underlying
		probability distribution. For a mixture model, this involves first
		sampling the component using the prior probabilities, and then sampling
		from the chosen distribution.


		Parameters
		----------
		n: int
			The number of samples to generate.
		

		Returns
		-------
		X: torch.tensor, shape=(n, self.d)
			Randomly generated samples.
		"""

		X = torch.zeros(n, self.d, dtype=torch.int32) - 1

		for i in range(self.d+1):
			for j, parents in enumerate(self._parents):
				if (X[0, j] != -1).item():
					continue

				if len(parents) == 0:
					X[:, j] = self.distributions[j].sample(n)[:, 0]
				else:
					X_ = X[:, parents].unsqueeze(-1)
					if (X_ == -1).any().item():
						continue

					X[:, j] = self.distributions[j].sample(n, X_)[:, 0]

		return X

	def log_probability(self, X):
		"""Calculate the log probability of each example.

		This method calculates the log probability of each example given the
		parameters of the distribution. The examples must be given in a 2D
		format.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate.

		Returns
		-------
		logp: torch.Tensor, shape=(-1,)
			The log probability of each example.
		"""

		X = _check_parameter(_cast_as_tensor(X), "X", ndim=2, 
			check_parameter=self.check_data)
		
		logps = torch.zeros(X.shape[0], device=X.device, dtype=torch.float32)
		for i, distribution in enumerate(self.distributions):
			parents = self._parents[i] + (i,)

			X_ = X[:, parents]
			if len(parents) > 1:
				X_ = X_.unsqueeze(-1)

			logps += distribution.log_probability(X_)

		return logps

	def predict(self, X):
		"""Infers the maximum likelihood value for each missing value.

		This method infers a probability distribution for each of the missing
		values in the data. It uses the factor graph representation of the
		Bayesian network to run the sum-product/loopy belief propagation
		algorithm. After the probability distribution is inferred, the maximum
		likeihood value for each variable is returned.

		The input to this method must be a torch.masked.MaskedTensor where the
		mask specifies which variables are observed (mask = True) and which ones
		are not observed (mask = False) for each of the values. When setting
		mask = False, it does not matter what the corresponding value in the
		tensor is. Different sets of variables can be observed or missing in
		different examples. 

		Unlike the `predict_proba` and `predict_log_proba` methods, this
		method preserves the dimensions of the original data because it does
		not matter how many categories a variable can take when you're only
		returning the maximally likely one.


		Parameters
		----------
		X: torch.masked.MaskedTensor, shape=(-1, d)
			The data to predict values for. The mask should correspond to
			whether the variable is observed in the example. 
		

		Returns
		-------
		y: torch.tensor, shape=(-1, d)
			A completed version of the incomplete input tensor. The missing
			variables are replaced with the maximally likely values from
			the sum-product algorithm, and observed variables are kept.
		"""

		y = [t.argmax(dim=1) for t in self.predict_proba(X)]
		return torch.vstack(y).T.contiguous()

	def predict_proba(self, X):
		"""Infers the probability of each category given the model and data.

		This method infers a probability distribution for each of the missing 
		values in the data. It uses the factor graph representation of the
		Bayesian network to run the sum-product/loopy belief propagation
		algorithm.

		The input to this method must be a torch.masked.MaskedTensor where the
		mask specifies which variables are observed (mask = True) and which ones
		are not observed (mask = False) for each of the values. When setting
		mask = False, it does not matter what the corresponding value in the
		tensor is. Different sets of variables can be observed or missing in
		different examples. 

		An important note is that, because each variable can have a different
		number of categories in the categorical setting, the return is a list
		of tensors where each element in that list is the marginal probability
		distribution for that variable. More concretely: the first element will
		be the distribution of values for the first variable across all
		examples. When the first variable has been provided as evidence, the
		distribution will be clamped to the value provided as evidence.

		..warning:: This inference is exact given a Bayesian network that has
		a tree-like structure, but is only approximate for other cases. When
		the network is acyclic, this procedure will converge, but if the graph
		contains cycles then there is no guarantee on convergence.


		Parameters
		----------
		X: torch.masked.MaskedTensor, shape=(-1, d)
			The data to predict values for. The mask should correspond to
			whether the variable is observed in the example. 
		

		Returns
		-------
		y: list of tensors, shape=(d,)
			A list of tensors where each tensor contains the distribution of
			values for that dimension.
		"""

		return self._factor_graph.predict_proba(X)

	def predict_log_proba(self, X):
		"""Infers the probability of each category given the model and data.

		This method is a wrapper around the `predict_proba` method and simply
		takes the log of each returned tensor.

		This method infers a log probability distribution for each of the 
		missing  values in the data. It uses the factor graph representation of 
		the Bayesian network to run the sum-product/loopy belief propagation
		algorithm.

		The input to this method must be a torch.masked.MaskedTensor where the
		mask specifies which variables are observed (mask = True) and which ones
		are not observed (mask = False) for each of the values. When setting
		mask = False, it does not matter what the corresponding value in the
		tensor is. Different sets of variables can be observed or missing in
		different examples. 

		An important note is that, because each variable can have a different
		number of categories in the categorical setting, the return is a list
		of tensors where each element in that list is the marginal probability
		distribution for that variable. More concretely: the first element will
		be the distribution of values for the first variable across all
		examples. When the first variable has been provided as evidence, the
		distribution will be clamped to the value provided as evidence.

		..warning:: This inference is exact given a Bayesian network that has
		a tree-like structure, but is only approximate for other cases. When
		the network is acyclic, this procedure will converge, but if the graph
		contains cycles then there is no guarantee on convergence.


		Parameters
		----------
		X: torch.masked.MaskedTensor, shape=(-1, d)
			The data to predict values for. The mask should correspond to
			whether the variable is observed in the example. 
		

		Returns
		-------
		y: list of tensors, shape=(d,)
			A list of tensors where each tensor contains the distribution of
			values for that dimension.
		"""

		return [torch.log(t) for t in self.predict_proba(X)]

	def fit(self, X, sample_weight=None):
		"""Fit the model to optionally weighted examples.

		This method will fit the provided distributions given the data and
		their weights. If a structure is provided as a set of edges, then
		this will use maximum likelihood estimates to fit each of those
		distributions. If no structure is provided, this will use the
		structure learning algorithm provided to jointly learn the structure
		and the parameters.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate. 

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.


		Returns
		-------
		self
		"""

		if self.algorithm is not None:
			self.structure = _learn_structure(X, sample_weight=sample_weight, 
				algorithm=self.algorithm, 
				include_parents=self.include_parents, 
				exclude_parents=self.exclude_parents, 
				max_parents=self.max_parents, pseudocount=self.pseudocount)

		if self.structure is not None:
			distributions = _from_structure(X, sample_weight, self.structure)
			self.add_distributions(distributions)

			for i, parents in enumerate(self.structure):
				if len(parents) > 0:
					for parent in parents:
						self.add_edge(distributions[parent], distributions[i])

		self.summarize(X, sample_weight=sample_weight)
		self.from_summaries()
		return self

	def summarize(self, X, sample_weight=None):
		"""Extract the sufficient statistics from a batch of data.

		This method calculates the sufficient statistics from optionally
		weighted data and adds them to the stored cache for each distribution
		in the network. Sample weights can either be provided as one
		value per example or as a 2D matrix of weights for each feature in
		each example.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, len, self.d)
			A set of examples to summarize.

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.


		Returns
		-------
		logp: torch.Tensor, shape=(-1,)
			The log probability of each example.
		"""

		if self.frozen:
			return

		if not self._initialized:
			self._initialize(len(X[0]))

		X, sample_weight = super().summarize(X, sample_weight=sample_weight)
		X = _check_parameter(X, "X", min_value=0, ndim=2, 
			check_parameter=self.check_data)

		for i, distribution in enumerate(self.distributions):
			parents = self._parents[i] + (i,)
			w = sample_weight[:, i]

			if len(parents) == 1:
				distribution.summarize(X[:, parents], sample_weight=w)
			else:
				distribution.summarize(X[:, parents].unsqueeze(-1), 
					sample_weight=w)

	def from_summaries(self):
		"""Update the model parameters given the extracted statistics.

		This method uses calculated statistics from calls to the `summarize`
		method to update the distribution parameters. Hyperparameters for the
		update are passed in at initialization time.

		Note: Internally, a call to `fit` is just a successive call to the
		`summarize` method followed by the `from_summaries` method.
		"""

		if self.frozen:
			return

		for distribution in self.distributions:
			distribution.from_summaries()

			if isinstance(distribution, ConditionalCategorical):
				p = torch.clone(distribution.probs[0])
				p /= torch.prod(torch.tensor(p.shape[:-1]))
				self._factor_mapping[distribution].probs = _cast_as_parameter(p)



#####

def _from_structure(X, sample_weight=None, structure=None, pseudocount=0.0):
	"""Fits a set of distributions to data given the structure.

	Given the structure, create the distribution objects and fit their 
	parameters to the given data. This does not perform structure learning.


	Parameters
	----------
	X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
		A set of examples to evaluate. 

	sample_weight: list, tuple, numpy.ndarray, torch.Tensor, or None
		A set of weights for the examples. This can be either of shape
		(-1, self.d) or a vector of shape (-1,). Default is ones.

	structure: tuple
		A tuple of tuples where the internal tuples indicate the parents of
		that variable. 

	pseudocount: double
		A pseudocount to add to each count. Default is 0.


	Returns
	-------
	model: pomegranate.bayesian_network.BayesianNetwork
		The fit Bayesian network.
	"""

	X = _check_parameter(_cast_as_tensor(X), "X", min_value=0, ndim=2,
		dtypes=(torch.int32, torch.int64))
	sample_weight = _check_parameter(_cast_as_tensor(sample_weight), "X", 
		min_value=0, ndim=1)

	n, d = X.shape

	if sample_weight is None:
		sample_weight = torch.ones(n, dtype=torch.float32, device=X.device)

	if structure is None:
		structure = tuple([tuple() for i in range(d)])

	model = BayesianNetwork()
	if X.device != model.device:
		model.to(X.device)

	d = len(structure)
	distributions = []

	for i, parents in enumerate(structure):
		if len(parents) == 0:
			d = Categorical()
			if d.device != X.device:
				d.to(X.device)

			d.fit(X[:, i:i+1], sample_weight=sample_weight)
		else:
			parents = parents + (i,)
			d = ConditionalCategorical()
			if d.device != X.device:
				d.to(X.device)

			d.fit(X[:, parents].unsqueeze(-1), sample_weight=sample_weight)

		distributions.append(d)

	return distributions


def _learn_structure(X, sample_weight=None, algorithm='chow-liu', 
	include_parents=None, exclude_parents=None, max_parents=None, 
	pseudocount=0, penalty=None, root=0):
	"""Learn the structure of a Bayesian network using data.

	This function will take in data, an algorithm, and parameters for that
	algorithm, and will return the learned structure. Currently supported
	algorithms are:

		- 'chow-liu': Learn a maximal spanning tree that is the most likely
			tree given the data and weights.
		- 'exact': A dynamic programming solution to the exact BNSL task that
			reduces the time from super-exponential to simply-exponential.


	Parameters
	----------
	X: torch.tensor, shape=(n, d)
		The data to fit the structure too, where each row is a sample and
		each column corresponds to the associated variable.
	
	sample_weight: torch.tensor, shape=(n,)
		The weight of each sample as a positive double. If None, all items are 
		weighted equally.

	algorithm: str
		The name of the algorithm to use for structure learning. Must be one
		of 'chow-liu' or 'exact'.
	
	include_parents: list or None
		A list of tuples where each inner tuple is made up of integers
		indicating the parents that must exist for that variable. For example,
		when passing in [(1,), (), ()], the first variable must have the second
		variable as a parent, and potentially others if learned, and the others
		do not have any parents that must be present. If None, no parents are
		forced. Default is None.

	exclude_parents: list or None
		A list of tuples where each inner tuple is made up of integers
		indicating the parents that cannot exist for that variable. For example,
		when passing in [(1,), (), ()], the first variable cannot have the 
		second variable as a parent, and the other variables have no
		restrictions on it. If None, no parents are excluded. Default is None.

	max_parents: int or None
		The maximum number of parents a variable can have. If used, this means
		using the k-learn procedure. Can drastically speed up algorithms.
		If None, no max on parents. Default is None.

	pseudocount: double
		A pseudocount to add to each count. Default is 0.
	
	penalty: float or None, optional
		The weighting of the model complexity term in the objective function.
		Increasing this value will encourage sparsity whereas setting the value
		to 0 will result in an unregularized structure. Default is
		log2(|D|) / 2 where |D| is the sum of the weights of the data.
	
	root: int
		When using a tree-based algorithm, like Chow-Liu, sets the variable
		that is the root of the tree.


	Returns
	-------
	structure: tuple, shape=(d,)
		A tuple of tuples, where each inner tuple represents a dimension in
		the data and contains the parents of that dimension.
	"""

	X = _check_parameter(_cast_as_tensor(X), "X", min_value=0, ndim=2,
		dtypes=(torch.int32, torch.int64))
	sample_weight = _check_parameter(_cast_as_tensor(sample_weight), "X", 
		min_value=0, ndim=1)

	if sample_weight is None:
		sample_weight = torch.ones(X.shape[0], dtype=torch.float32, 
			device=X.device)

	if algorithm == 'chow-liu':
		structure = _categorical_chow_liu(X, sample_weight=sample_weight, 
			pseudocount=pseudocount, root=root)
	elif algorithm == 'exact':
		structure = _categorical_exact(X, sample_weight=sample_weight, 
			include_parents=include_parents, exclude_parents=exclude_parents, 
			max_parents=max_parents, pseudocount=pseudocount)

	return structure


def _categorical_chow_liu(X, sample_weight=None, pseudocount=0.0, root=0):
	"""An internal function for calculating a Chow-Liu tree.

	This function calculates a Chow-Liu tree on categorical data which is
	potentially weighted. A Chow-Liu tree is essentially a maximum spanning
	tree on the information content that adding each new variable might
	add.


	Parameters
	----------
	X: torch.tensor, shape=(n, d)
		The data to fit the structure too, where each row is a sample and
		each column corresponds to the associated variable.
	
	sample_weight: torch.tensor or None, shape=(n,)
		The weight of each sample as a positive double. If None, all items are 
		weighted equally.

	pseudocount: double
		A pseudocount to add to each count. Default is 0.
	
	root: int
		When using a tree-based algorithm, like Chow-Liu, sets the variable
		that is the root of the tree.


	Returns
	-------
	structure: tuple
		A tuple of tuples where the internal tuples indicate the parents of
		that variable. 
	"""

	X = _check_parameter(_cast_as_tensor(X), "X", min_value=0, ndim=2,
		dtypes=(torch.int32, torch.int64))
	sample_weight = _check_parameter(_cast_as_tensor(sample_weight), "X", 
		min_value=0, ndim=1)

	pseudocount = _check_parameter(pseudocount, "pseudocount", min_value=0,
		ndim=0)
	root = _check_parameter(root, "root", min_value=0, max_value=X.shape[1],
		dtypes=(int, numpy.int32, numpy.int64, torch.int32, torch.int64))

	n, d = X.shape

	if sample_weight is None:
		sample_weight = torch.ones(n)

	start = time.time()
	n_categories = tuple(x.item() for x in X.max(dim=0).values + 1)
	max_categories = max(n_categories)

	c = max_categories ** 2
	dtype = sample_weight.dtype
	count = torch.empty(c, d, dtype=dtype, device=X.device)
	mutual_info = torch.zeros(d, d, dtype=dtype, device=X.device)
	
	marginals_i = torch.empty(d, c, dtype=dtype, device=X.device)
	marginals_r = torch.empty(d, c, dtype=dtype, device=X.device)

	for j in range(d):
		marg = torch.zeros(max_categories, dtype=dtype, device=X.device)
		marg.scatter_add_(0, X[:, j], sample_weight)

		marginals_i[j] = marg.repeat_interleave(max_categories)
		marginals_r[j] = marg.repeat(max_categories)

	w = sample_weight.unsqueeze(-1).expand(-1, d)

	for j in range(d):
		X_j = X[:, j:j+1] * max_categories + X
		m = marginals_i[j:j+1] * marginals_r

		count[:] = pseudocount
		count.scatter_add_(0, X_j, w)

		mutual_info[j] -= torch.sum(count * torch.log(count / m.T), dim=0)
		mutual_info[:, j] = mutual_info[j] 

	structure = [[] for i in range(d)]
	visited = [root]
	unvisited = list(range(d))
	unvisited.remove(root)

	for i in range(d-1):
		min_score = float("inf")
		min_x = -1
		min_y = -1

		idx = mutual_info[visited][:, unvisited].argmin()
		row, col = (idx // len(unvisited)).item(), (idx % len(unvisited)).item()
		min_x, min_y = visited[row], unvisited[col]

		structure[min_y].append(min_x)
		visited.append(min_y)
		unvisited.remove(min_y)

	return tuple(tuple(x) for x in structure)


def _categorical_exact(X, sample_weight=None, include_parents=None, 
	exclude_parents=None, pseudocount=0, penalty=None, max_parents=None):
	"""Find the optimal graph over a set of variables with no other knowledge.
	
	This is the naive dynamic programming structure learning task where the
	optimal graph is identified from a set of variables using an order graph
	and parent graphs. This can be used either when no constraint graph is
	provided or for a SCC which is made up of a node containing a self-loop.
	This is a reference implementation that uses the naive shortest path
	algorithm over the entire order graph. The 'exact' option uses the A* path
	in order to avoid considering the full order graph.
	

	Parameters
	----------
	X: numpy.ndarray, shape=(n, d)
		The data to fit the structure too, where each row is a sample and
		each column corresponds to the associated variable.
	
	sample_weight: numpy.ndarray, shape=(n,)
		The weight of each sample as a positive double. Default is None.
	
	include_parents: list or None
		A set of (parent, child) tuples where each tuple is an edge that
		must exist in the found structure.
	
	exclude_parents: list or None
		A set of (parent, child) tuples where each tuple is an edge that
		cannot exist in the found structure.
	
	penalty: float or None, optional
		The weighting of the model complexity term in the objective function.
		Increasing this value will encourage sparsity whereas setting the value
		to 0 will result in an unregularized structure. Default is
		log2(|D|) / 2 where |D| is the sum of the weights of the data.
	
	max_parents: int
		The maximum number of parents a node can have. If used, this means
		using the k-learn procedure. Can drastically speed up algorithms.
		If -1, no max on parents. Default is -1.
		

	Returns
	-------
	structure: tuple, shape=(d,)
		The parents for each variable in this SCC
	"""

	X = _check_parameter(_cast_as_tensor(X), "X", min_value=0, ndim=2,
		dtypes=(torch.int32, torch.int64))
	sample_weight = _check_parameter(_cast_as_tensor(sample_weight), "X", 
		min_value=0, ndim=1)

	n, d = X.shape
	n_categories = X.max(axis=0).values + 1

	if sample_weight is None:
		sample_weight = torch.ones(n)

	w = sample_weight.sum()
	if max_parents is None:
		max_parents = int(torch.log2(2 * w / torch.log2(w)).item())

	parent_graphs = []
	for i in range(d):
		exclude = None if exclude_parents is None else exclude_parents[i]
		parent_set = tuple(set(range(d)) - set([i]))

		parent_graph = _generate_parent_graph(X, sample_weight=sample_weight, 
			n_categories=n_categories, column_idx=i, 
			include_parents=include_parents, exclude_parents=exclude,
			pseudocount=pseudocount, penalty=penalty, max_parents=max_parents, 
			parent_set=parent_set) 
		parent_graphs.append(parent_graph)

	order_graph = nx.DiGraph()

	for i in range(d+1):
		for subset in itertools.combinations(range(d), i):
			order_graph.add_node(subset)

			for variable in subset:
				parent = tuple(v for v in subset if v != variable)

				structure, weight = parent_graphs[variable][parent]
				weight = -weight if weight < 0 else 0
				order_graph.add_edge(parent, subset, weight=weight,
					structure=structure)

	path = sorted(nx.all_shortest_paths(order_graph, source=(),
		target=tuple(range(d)), weight="weight"))[0]

	score, structure = 0, list( None for i in range(d) )
	for u, v in zip(path[:-1], path[1:]):
		idx = list(set(v) - set(u))[0]
		parents = order_graph.get_edge_data(u, v)['structure']
		structure[idx] = parents
		score -= order_graph.get_edge_data(u, v)['weight']

	return tuple(structure)


def _generate_parent_graph(X, sample_weight, n_categories, column_idx, 
	include_parents=None, exclude_parents=None, pseudocount=0, penalty=None, 
	max_parents=None, parent_set=None):
	"""Generate a parent graph for a single variable over its parents.

	This will generate the parent graph for a single parents given the data.
	A parent graph is the dynamically generated best parent set and respective
	score for each combination of parent variables. For example, if we are
	generating a parent graph for x1 over x2, x3, and x4, we may calculate that
	having x2 as a parent is better than x2,x3 and so store the value
	of x2 in the node for x2,x3.
	
	Parameters
	----------
	X: list, numpy.ndarray, torch.tensor, shape=(n, d)
		The data to fit the structure too, where each row is a sample and
		each column corresponds to the associated variable.
	
	weights: list, numpy.ndarray, torch.tensor, shape=(n,)
		The weight of each sample as a positive double. Default is None.
	
	n_categories: list, numpy.ndarray, torch.tensor, shape=(d,)
		The number of unique keys in each column.
	
	column_idx: int
		The column index to build the parent graph for.
	
	include_parents: list or tuple or None
		A set of integers indicating the parents that must be included in the
		returned structure. Default is None.
	
	exclude_parents: list or tuple or None
		A set of integers indicating the parents that must be excluded from
		the returned structure. Default is None.

	pseudocount: double
		A pseudocount to add to each count. Default is 0.
	
	penalty: float or None, optional
		The weighting of the model complexity term in the objective function.
		Increasing this value will encourage sparsity whereas setting the value
		to 0 will result in an unregularized structure. Default is
		log2(|D|) / 2 where |D| is the sum of the weights of the data.
	
	max_parents: int
		The maximum number of parents a node can have. If used, this means
		using the k-learn procedure. Can drastically speed up algorithms.
		If -1, no max on parents. Default is -1.
	
	parent_set: tuple, default ()
		The variables which are possible parents for this variable. By default,
		this should be all variables except for idx. If excluded parents are
		passed in, this should exclude those variables as well.
	

	Returns
	-------
	structure: tuple, shape=(d,)
		The parents for each variable in this SCC
	"""

	n, d = X.shape
	n_categories = n_categories.numpy()
	max_n_categories = int(n_categories.max())

	parent_graph = {}
	X_dicts = {}
	X_cols = X.T.contiguous().type(torch.int64)

	w = sample_weight.sum()
	log_w = torch.log2(sample_weight.sum()).item() / 2

	if max_parents is None:
		max_parents = int(torch.log2(2 * w / torch.log2(w)).item())

	if parent_set is None:
		parent_set = tuple(set(range(d)) - set([column_idx]))

	if include_parents is None:
		include_parents = []

	if exclude_parents is None:
		exclude_parents = set()
	else:
		exclude_parents = set(exclude_parents)

	for j in range(len(parent_set)+1):
		c_dims = tuple([max_n_categories for _ in range(j+1)])
		counts = torch.zeros(*c_dims, dtype=sample_weight.dtype)

		offset = max_n_categories ** j
		X_cols_ = X_cols * offset

		for subset in itertools.combinations(parent_set, j):
			best_structure = ()
			best_score = float("-inf")

			if j <= max_parents:
				for parent in include_parents:
					if parent not in subset:
						break
				else:
					for var in subset:
						if var in exclude_parents:
							break
					else:
						if j  == 0:
							X_idxs = torch.zeros(n, dtype=torch.int64)
							idx = column_idx
						else:
							X_idxs = X_dicts[subset[:-1]]
							idx = subset[-1]

						n_params = (numpy.prod(n_categories[list(subset)]) * 
							(n_categories[column_idx] - 1))

						X_idxs = torch.clone(X_idxs) + X_cols_[idx]

						counts[:] = pseudocount
						counts.view(-1).scatter_add_(0, X_idxs, sample_weight)
						marginal_counts = counts.sum(dim=-1, keepdims=True)

						logp = torch.sum(counts * torch.log2(counts / 
							marginal_counts))

						best_structure = subset
						best_score = logp - log_w * n_params
						X_dicts[subset] = X_idxs

			for k, variable in enumerate(subset):
				parent_subset = tuple(l for l in subset if l != variable)
				structure, score = parent_graph[parent_subset]

				if score > best_score:
					best_score = score
					best_structure = structure

			parent_graph[subset] = (best_structure, best_score)

	return parent_graph



# _distribution.py
# Jacob Schreiber <jmschreiber91@gmail.com>

import torch

from .._utils import _cast_as_tensor
from .._utils import _update_parameter
from .._utils import _cast_as_parameter
from .._utils import _check_parameter
from .._utils import _reshape_weights


class Distribution(torch.nn.Module):
	"""A base distribution object.

	This distribution is inherited by all the other distributions.
	"""

	def __init__(self, inertia, frozen, check_data):
		super(Distribution, self).__init__()
		self._device = _cast_as_parameter([0.0])

		_check_parameter(inertia, "inertia", min_value=0, max_value=1, ndim=0)
		_check_parameter(frozen, "frozen", value_set=[True, False], ndim=0)
		_check_parameter(check_data, "check_data", value_set=[True, False],
			ndim=0)

		self.register_buffer("inertia", _cast_as_tensor(inertia))
		self.register_buffer("frozen", _cast_as_tensor(frozen))
		self.register_buffer("check_data", _cast_as_tensor(check_data))

		self._initialized = False

	@property
	def device(self):
		try:
			return next(self.parameters()).device
		except:
			return 'cpu'

	@property
	def dtype(self):
		return next(self.parameters()).dtype

	def freeze(self):
		self.register_buffer("frozen", _cast_as_tensor(True))
		return self

	def unfreeze(self):
		self.register_buffer("frozen", _cast_as_tensor(False))
		return self

	def forward(self, X):
		self.summarize(X)
		return self.log_probability(X)

	def backward(self, X):
		self.from_summaries()
		return X

	def _initialize(self, d):
		self.d = d
		self._reset_cache()

	def _reset_cache(self):
		raise NotImplementedError

	def probability(self, X):
		return torch.exp(self.log_probability(X))

	def log_probability(self, X):
		raise NotImplementedError

	def fit(self, X, sample_weight=None):
		self.summarize(X, sample_weight=sample_weight)
		self.from_summaries()
		return self

	def summarize(self, X, sample_weight=None):
		if not self._initialized:
			self._initialize(len(X[0]))

		X = _cast_as_tensor(X)
		_check_parameter(X, "X", ndim=2, shape=(-1, self.d), 
			check_parameter=self.check_data)

		sample_weight = _reshape_weights(X, _cast_as_tensor(sample_weight), 
			device=self.device)

		return X, sample_weight

	def from_summaries(self):
		raise NotImplementedError


class ConditionalDistribution(Distribution):
	def __init__(self, inertia, frozen, check_data):
		super().__init__(inertia=inertia, frozen=frozen, check_data=check_data)

	def marginal(self, dim):
		raise NotImplementedError


# test_bayes_classifier.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import os
import numpy
import torch
import pytest

from pomegranate.gmm import GeneralMixtureModel
from pomegranate.hmm import DenseHMM
from pomegranate.distributions import Exponential

from .tools import assert_raises
from numpy.testing import assert_array_equal
from numpy.testing import assert_array_almost_equal


inf = float("inf")


@pytest.fixture
def X():
	return [[1, 2, 0],
	     [0, 0, 1],
	     [1, 1, 2],
	     [2, 2, 2],
	     [3, 1, 0],
	     [5, 1, 4],
	     [2, 1, 0],
	     [1, 0, 2],
	     [1, 1, 0],
	     [0, 2, 1],
	     [0, 0, 0]]


@pytest.fixture
def priors():
	return torch.tensor(numpy.array([
		[0.5, 0.5],
	    [0.5, 0.5],
	    [0.5, 0.5],
	    [0.0, 1.0],
	    [0.5, 0.5],
	    [0.3, 0.7],
	    [0.6, 0.4],
	    [0.5, 0.5],
	    [0.0, 1.0],
	    [1.0, 0.0],
	    [0.5, 0.5]
	]))


@pytest.fixture
def gmm():
	d = [Exponential([2.1, 0.3, 0.1]), Exponential([1.5, 3.1, 2.2])]
	return GeneralMixtureModel(d, priors=[0.7, 0.3])


@pytest.fixture
def hmm():
	d = [Exponential([2.1, 0.3, 0.1]), Exponential([1.5, 3.1, 2.2])]
	return DenseHMM(d, edges=[[0.8, 0.2], [0.4, 0.6]], starts=[0.4, 0.6],
		ends=[0.5, 0.5])


###


def _test_gmm_raises(func, X, priors):
	X = numpy.array(X, dtype='float32')
	priors = numpy.array(priors, dtype='float32')

	assert_raises(ValueError, func, X, priors+1)
	assert_raises(ValueError, func, X, priors-1)
	assert_raises(ValueError, func, X, priors/2.0)
	assert_raises(ValueError, func, X, priors[:5])
	assert_raises(ValueError, func, X[:5], priors)
	assert_raises(ValueError, func, X, priors[:,0])
	assert_raises(ValueError, func, X, priors[:, :1])


def test_gmm_emission_matrix(gmm, X, priors):
	y_hat = gmm._emission_matrix(X)
	assert_array_almost_equal(y_hat, [
		[ -4.7349,  -4.8411],
        [ -7.5921,  -3.9838],
        [-21.4016,  -5.4276],
        [-25.2111,  -6.4169],
        [ -2.3540,  -5.8519],
        [-43.3063,  -9.0034],
        [ -1.8778,  -5.1852],
        [-18.0682,  -5.1051],
        [ -1.4016,  -4.5185],
        [-14.2587,  -4.6290],
        [  2.4079,  -3.5293]], 4)

	y_hat = gmm._emission_matrix(X, priors=priors)
	assert_array_almost_equal(y_hat, [
		[ -5.4281,  -5.5343],
        [ -8.2852,  -4.6770],
        [-22.0947,  -6.1208],
        [    -inf,  -6.4169],
        [ -3.0471,  -6.5450],
        [-44.5103,  -9.3601],
        [ -2.3886,  -6.1015],
        [-18.7614,  -5.7982],
        [    -inf,  -4.5185],
        [-14.2587,     -inf],
        [  1.7148,  -4.2224]], 4)


def test_gmm_emission_matrix_raises(gmm, X, priors):
	_test_gmm_raises(gmm._emission_matrix, X, priors)


def test_gmm_probability(gmm, X, priors):
	y_hat = gmm.probability(X)
	assert_array_almost_equal(y_hat, numpy.exp([-4.0935, -3.9571, -5.4276, 
		-6.4169, -2.3241, -9.0034, -1.8418, -5.1051, -1.3582, -4.6289,  
		2.4106]), 3)

	y_hat = gmm.probability(X, priors=priors)
	assert_array_almost_equal(y_hat, [8.3407e-03, 9.5592e-03, 2.1967e-03, 
		1.6337e-03, 4.8933e-02, 8.6094e-05, 9.3998e-02, 3.0330e-03, 
		1.0905e-02, 6.4197e-07, 5.5702e+00], 3)	


def test_gmm_probability_raises(gmm, X, priors):
	_test_gmm_raises(gmm.probability, X, priors)


def test_gmm_log_probability(gmm, X, priors):
	y_hat = gmm.log_probability(X)
	assert_array_almost_equal(y_hat, [-4.0935, -3.9571, -5.4276, -6.4169, 
		-2.3241, -9.0034, -1.8418, -5.1051, -1.3582, -4.6289,  2.4106], 4)

	y_hat = gmm.log_probability(X, priors=priors)
	assert_array_almost_equal(y_hat, [-4.7866, -4.6503, -6.1208, -6.4169,
		-3.0173, -9.3601, -2.3645, -5.7982, -4.5185, -14.2587, 1.7174], 4)	


def test_gmm_log_probability_raises(gmm, X, priors):
	_test_gmm_raises(gmm.log_probability, X, priors)


def test_gmm_predict(gmm, X, priors):
	y_hat = gmm.predict(X)
	assert_array_equal(y_hat, [0, 1, 1, 1, 0, 1, 0, 1, 0, 1, 0])

	y_hat = gmm.predict(X, priors=priors)
	assert_array_equal(y_hat, [0, 1, 1, 1, 0, 1, 0, 1, 1, 0, 0])	


def test_gmm_predict_raises(gmm, X, priors):
	_test_gmm_raises(gmm.predict, X, priors)


def test_gmm_predict_proba(gmm, X, priors):
	y_hat = gmm.predict_proba(X)
	assert_array_almost_equal(y_hat, [
		[5.2653e-01, 4.7347e-01],
        [2.6385e-02, 9.7361e-01],
        [1.1551e-07, 1.0000e+00],
        [6.8830e-09, 1.0000e+00],
        [9.7063e-01, 2.9372e-02],
        [1.2660e-15, 1.0000e+00],
        [9.6468e-01, 3.5317e-02],
        [2.3451e-06, 1.0000e+00],
        [9.5759e-01, 4.2413e-02],
        [6.5741e-05, 9.9993e-01],
        [9.9737e-01, 2.6323e-03]], 4)

	y_hat = gmm.predict_proba(X, priors=priors)
	assert_array_almost_equal(y_hat, [
		[5.2653e-01, 4.7347e-01],
        [2.6385e-02, 9.7361e-01],
        [1.1551e-07, 1.0000e+00],
        [0.0000e+00, 1.0000e+00],
        [9.7063e-01, 2.9372e-02],
        [5.4256e-16, 1.0000e+00],
        [9.7618e-01, 2.3825e-02],
        [2.3451e-06, 1.0000e+00],
        [0.0000e+00, 1.0000e+00],
        [1.0000e+00, 0.0000e+00],
        [9.9737e-01, 2.6323e-03]], 4)	


def test_gmm_predict_proba_raises(gmm, X, priors):
	_test_gmm_raises(gmm.predict_proba, X, priors)


def test_gmm_predict_log_proba(gmm, X, priors):
	y_hat = gmm.predict_log_proba(X)
	assert_array_almost_equal(y_hat, [
		[-6.4145e-01, -7.4766e-01],
        [-3.6350e+00, -2.6740e-02],
        [-1.5974e+01,  0.0000e+00],
        [-1.8794e+01,  0.0000e+00],
        [-2.9812e-02, -3.5277e+00],
        [-3.4303e+01,  0.0000e+00],
        [-3.5955e-02, -3.3434e+00],
        [-1.2963e+01, -2.3842e-06],
        [-4.3338e-02, -3.1603e+00],
        [-9.6298e+00, -6.5804e-05],
        [-2.6357e-03, -5.9399e+00]], 3)

	y_hat = gmm.predict_log_proba(X, priors=priors)
	assert_array_almost_equal(y_hat, [
		[-6.4145e-01, -7.4766e-01],
        [-3.6350e+00, -2.6740e-02],
        [-1.5974e+01,  0.0000e+00],
        [       -inf,  0.0000e+00],
        [-2.9812e-02, -3.5277e+00],
        [-3.5150e+01,  0.0000e+00],
        [-2.4113e-02, -3.7370e+00],
        [-1.2963e+01, -2.3842e-06],
        [       -inf,  0.0000e+00],
        [ 0.0000e+00,        -inf],
        [-2.6358e-03, -5.9399e+00]], 3)	


def test_gmm_predict_log_proba_raises(gmm, X, priors):
	_test_gmm_raises(gmm.predict_log_proba, X, priors)


def test_gmm_summarize(gmm, X, priors):
	gmm.summarize(X, priors=priors)

	assert_array_almost_equal(gmm._w_sum, [4.497087, 6.502913], 4)

	assert_array_almost_equal(gmm.distributions[0]._w_sum, [4.497087, 4.497087, 
		4.497087], 4)
	assert_array_almost_equal(gmm.distributions[0]._xw_sum, [5.390766, 4.999861,
		1.02639], 4)

	assert_array_almost_equal(gmm.distributions[1]._w_sum, [6.502912, 6.502912, 
		6.502912], 4)
	assert_array_almost_equal(gmm.distributions[1]._xw_sum, [10.609234, 
		6.000139, 10.97361], 4)


def test_gmm_summarize_raises(gmm, X, priors):
	X = numpy.array(X, dtype='float32')
	priors = numpy.array(priors, dtype='float32')
	func = gmm.summarize

	assert_raises(ValueError, func, X, None, priors+1)
	assert_raises(ValueError, func, X, None, priors-1)
	assert_raises(ValueError, func, X, None, priors/2.0)
	assert_raises(ValueError, func, X, None, priors[:5])
	assert_raises(ValueError, func, X[:5], None, priors)
	assert_raises(ValueError, func, X, None, priors[:,0])
	assert_raises(ValueError, func, X, None, priors[:, :1])


def test_gmm_fit(X, priors):
	X = numpy.array(X) + 1

	d1 = Exponential([2.1, 0.3, 0.1])
	d2 = Exponential([1.5, 3.1, 2.2])
	gmm = GeneralMixtureModel([d1, d2], priors=[0.7, 0.3])
	gmm.fit(X)

	assert_array_almost_equal(gmm._log_priors, [-8.9908e+00, -1.2458e-04], 4)	
	assert_array_almost_equal(d1.scales, [1.7997, 1.7256, 1.5104], 4)
	assert_array_almost_equal(d2.scales, [2.4546, 2.    , 2.091], 4)


	d1 = Exponential([2.1, 0.3, 0.1])
	d2 = Exponential([1.5, 3.1, 2.2])
	gmm = GeneralMixtureModel([d1, d2], priors=[0.7, 0.3])
	gmm.fit(X, priors=priors)

	assert_array_almost_equal(gmm._log_priors, [-1.607879, -0.223534], 4)	
	assert_array_almost_equal(d1.scales, [1.5753, 2.2913, 1.8507], 4)
	assert_array_almost_equal(d2.scales, [2.6748, 1.927 , 2.1511], 4)


def test_gmm_fit_raises(gmm, X, priors):
	X = numpy.array(X, dtype='float32')
	priors = numpy.array(priors, dtype='float32')
	func = gmm.fit

	assert_raises(ValueError, func, X, None, priors+1)
	assert_raises(ValueError, func, X, None, priors-1)
	assert_raises(ValueError, func, X, None, priors/2.0)
	assert_raises(ValueError, func, X, None, priors[:5])
	assert_raises(ValueError, func, X[:5], None, priors)
	assert_raises(ValueError, func, X, None, priors[:,0])
	assert_raises(ValueError, func, X, None, priors[:, :1])



#


def _test_hmm_raises(func, X, priors):
	X = numpy.array(X, dtype='float32')
	priors = numpy.array(priors, dtype='float32')[None, :, :]

	assert_raises(ValueError, func, X, None, priors+1)
	assert_raises(ValueError, func, X, None, priors-1)
	assert_raises(ValueError, func, X, None, priors/2.0)
	assert_raises(ValueError, func, X, None, priors[:, :5])
	assert_raises(ValueError, func, X[:, :5], None, priors)
	assert_raises(ValueError, func, X, None, priors[:, :, 0])
	assert_raises(ValueError, func, X, None, priors[:, :, :1])


def test_hmm_emission_matrix(hmm, X, priors):
	X = numpy.array([X])
	priors = priors.unsqueeze(0)

	y_hat = hmm._emission_matrix(X)
	assert_array_almost_equal(y_hat, [
		[[ -4.3782,  -3.6372],
         [ -7.2354,  -2.7799],
         [-21.0449,  -4.2237],
         [-24.8544,  -5.2129],
         [ -1.9973,  -4.6479],
         [-42.9497,  -7.7994],
         [ -1.5211,  -3.9812],
         [-17.7116,  -3.9011],
         [ -1.0449,  -3.3146],
         [-13.9020,  -3.4250],
         [  2.7646,  -2.3253]]], 4)

	y_hat = hmm._emission_matrix(X, priors=priors)
	assert_array_almost_equal(y_hat, [
		[[ -5.0714,  -4.3303],
         [ -7.9285,  -3.4730],
         [-21.7381,  -4.9168],
         [    -inf,  -5.2129],
         [ -2.6904,  -5.3411],
         [-44.1536,  -8.1561],
         [ -2.0319,  -4.8975],
         [-18.4047,  -4.5942],
         [    -inf,  -3.3146],
         [-13.9020,     -inf],
         [  2.0715,  -3.0185]]], 4)


def test_hmm_emission_matrix_raises(hmm, X, priors):
	X = numpy.array(X, dtype='float32')
	priors = numpy.array(priors, dtype='float32')[None, :, :]
	func = hmm._emission_matrix

	assert_raises(ValueError, func, X, priors+1)
	assert_raises(ValueError, func, X, priors-1)
	assert_raises(ValueError, func, X, priors/2.0)
	assert_raises(ValueError, func, X, priors[:, :5])
	assert_raises(ValueError, func, X[:, :5], priors)
	assert_raises(ValueError, func, X, priors[:, :, 0])
	assert_raises(ValueError, func, X, priors[:, :, :1])


def test_hmm_forward(hmm, X, priors):
	X = numpy.array([X])
	priors = priors.unsqueeze(0)

	y_hat = hmm.forward(X)
	assert_array_almost_equal(y_hat, [
		[[ -5.2945,  -4.1480],
         [-11.8077,  -7.3380],
         [-29.2766, -12.0687],
         [-37.8394, -17.7924],
         [-20.7060, -22.9511],
         [-63.8272, -29.8389],
         [-32.2763, -34.3310],
         [-50.1489, -37.4616],
         [-39.4228, -41.2870],
         [-53.4733, -44.0753],
         [-42.2268, -46.9115]]], 4)

	y_hat = hmm.forward(X, priors=priors)
	assert_array_almost_equal(y_hat, [
		[[ -5.9877,  -4.8411],
         [-13.1940,  -8.7243],
         [-31.3560, -14.1481],
         [    -inf, -19.8719],
         [-23.4786, -25.7237],
         [-67.8038, -32.9682],
         [-35.9164, -38.3766],
         [-54.5024, -41.8919],
         [    -inf, -45.7173],
         [-60.5357,     -inf],
         [-58.6873, -65.1636]]], 4)


def test_hmm_forward_raises(hmm, X, priors):
	_test_hmm_raises(hmm.forward, X, priors)


def test_hmm_backward(hmm, X, priors):
	X = numpy.array([X])
	priors = priors.unsqueeze(0)

	y_hat = hmm.backward(X)
	assert_array_almost_equal(y_hat, [
		[[-39.9505, -38.8647],
         [-36.6752, -35.5766],
         [-31.9407, -30.8421],
         [-24.6495, -25.1184],
         [-22.4807, -21.3821],
         [-12.6419, -13.0719],
         [-10.9597,  -9.8611],
         [ -5.0633,  -5.4492],
         [ -3.8699,  -2.7714],
         [  1.8499,   1.1644],
         [ -0.6931,  -0.6931]]], 4)

	y_hat = hmm.backward(X, priors=priors)
	assert_array_almost_equal(y_hat, [
		[[-55.7255, -54.6397],
         [-51.7571, -50.6585],
         [-46.3294, -45.2308],
         [-39.0382, -39.5071],
         [-36.1763, -35.0777],
         [-25.9039, -26.4107],
         [-23.6907, -22.5921],
         [-18.5856, -17.4870],
         [-12.9685, -13.6616],
         [  1.1567,   0.4712],
         [ -0.6931,  -0.6931]]], 4)


def test_hmm_backward_raises(hmm, X, priors):
	_test_hmm_raises(hmm.backward, X, priors)


def test_hmm_forward_backward(hmm, X, priors):
	X = numpy.array([X])
	priors = priors.unsqueeze(0)

	t, r, starts, ends, logp = hmm.forward_backward(X)
	assert_array_almost_equal(t, [[[1.8028e-03, 2.2629e+00],
         [3.1568e+00, 4.5785e+00]]], 4)

	assert_array_almost_equal(r, [
		[[-2.3343e+00, -1.0190e-01],
         [-5.5721e+00, -3.8109e-03],
         [-1.8306e+01,  0.0000e+00],
         [-1.9578e+01,  0.0000e+00],
         [-2.7591e-01, -1.4225e+00],
         [-3.3558e+01,  0.0000e+00],
         [-3.2527e-01, -1.2813e+00],
         [-1.2301e+01, -3.8147e-06],
         [-3.8184e-01, -1.1476e+00],
         [-8.7126e+00, -1.6403e-04],
         [-9.1934e-03, -4.6938e+00]]], 3)

	assert_array_almost_equal(starts, [[0.0969, 0.9031]], 4)
	assert_array_almost_equal(ends, [[0.9908, 0.0092]], 4)
	assert_array_almost_equal(logp, [-42.9108], 4)

	t, r, starts, ends, logp = hmm.forward_backward(X, priors=priors)
	assert_array_almost_equal(t, [[[0.9999, 1.6556], [2.5572, 4.7872]]], 4)

	assert_array_almost_equal(r, [
		[[-2.3343e+00, -1.0190e-01],
         [-5.5721e+00, -3.8109e-03],
         [-1.8306e+01,  0.0000e+00],
         [       -inf,  0.0000e+00],
         [-2.7591e-01, -1.4225e+00],
         [-3.4329e+01,  0.0000e+00],
         [-2.2815e-01, -1.5897e+00],
         [-1.3709e+01,  0.0000e+00],
         [       -inf,  0.0000e+00],
         [ 0.0000e+00,        -inf],
         [-1.5373e-03, -6.4778e+00]]], 3)

	assert_array_almost_equal(starts, [[0.0969, 0.9031]], 4)
	assert_array_almost_equal(ends, [[0.9985, 0.0015]], 4)
	assert_array_almost_equal(logp, [-59.3789], 4)


def test_hmm_forward_backward_raises(hmm, X, priors):
	_test_hmm_raises(hmm.forward_backward, X, priors)


def test_hmm_probability(hmm, X, priors):
	X = numpy.array([X])
	priors = priors.unsqueeze(0)

	y_hat = hmm.probability(X)
	assert_array_almost_equal(y_hat, [2.3125e-19], 4)

	y_hat = hmm.probability(X, priors=priors)
	assert_array_almost_equal(y_hat, [1.6295e-26], 4)


def test_hmm_probability_raises(hmm, X, priors):
	X = numpy.array(X, dtype='float32')
	priors = numpy.array(priors, dtype='float32')[None, :, :]
	func = hmm.probability

	assert_raises(ValueError, func, X, priors+1)
	assert_raises(ValueError, func, X, priors-1)
	assert_raises(ValueError, func, X, priors/2.0)
	assert_raises(ValueError, func, X, priors[:, :5])
	assert_raises(ValueError, func, X[:, :5], priors)
	assert_raises(ValueError, func, X, priors[:, :, 0])
	assert_raises(ValueError, func, X, priors[:, :, :1])


def test_hmm_log_probability(hmm, X, priors):
	X = numpy.array([X])
	priors = priors.unsqueeze(0)

	y_hat = hmm.log_probability(X)
	assert_array_almost_equal(y_hat, [-42.9108], 4)

	y_hat = hmm.log_probability(X, priors=priors)
	assert_array_almost_equal(y_hat, [-59.3789], 4)


def test_hmm_log_probability_raises(hmm, X, priors):
	X = numpy.array(X, dtype='float32')
	priors = numpy.array(priors, dtype='float32')[None, :, :]
	func = hmm.log_probability

	assert_raises(ValueError, func, X, priors+1)
	assert_raises(ValueError, func, X, priors-1)
	assert_raises(ValueError, func, X, priors/2.0)
	assert_raises(ValueError, func, X, priors[:, :5])
	assert_raises(ValueError, func, X[:, :5], priors)
	assert_raises(ValueError, func, X, priors[:, :, 0])
	assert_raises(ValueError, func, X, priors[:, :, :1])


def test_hmm_predict(hmm, X, priors):
	X = numpy.array([X])
	priors = priors.unsqueeze(0)

	y_hat = hmm.predict(X)
	assert_array_almost_equal(y_hat, [[1, 1, 1, 1, 0, 1, 0, 1, 0, 1, 0]], 4)

	y_hat = hmm.predict(X, priors=priors)
	assert_array_almost_equal(y_hat, [[1, 1, 1, 1, 0, 1, 0, 1, 1, 0, 0]], 4)


def test_hmm_predict_raises(hmm, X, priors):
	X = numpy.array(X, dtype='float32')
	priors = numpy.array(priors, dtype='float32')[None, :, :]
	func = hmm.predict

	assert_raises(ValueError, func, X, priors+1)
	assert_raises(ValueError, func, X, priors-1)
	assert_raises(ValueError, func, X, priors/2.0)
	assert_raises(ValueError, func, X, priors[:, :5])
	assert_raises(ValueError, func, X[:, :5], priors)
	assert_raises(ValueError, func, X, priors[:, :, 0])
	assert_raises(ValueError, func, X, priors[:, :, :1])


def test_hmm_predict_proba(hmm, X, priors):
	X = numpy.array([X])
	priors = priors.unsqueeze(0)

	y_hat = hmm.predict_proba(X)
	assert_array_almost_equal(y_hat, [
		[[9.6881e-02, 9.0312e-01],
         [3.8024e-03, 9.9620e-01],
         [1.1210e-08, 1.0000e+00],
         [3.1428e-09, 1.0000e+00],
         [7.5888e-01, 2.4112e-01],
         [2.6658e-15, 1.0000e+00],
         [7.2233e-01, 2.7767e-01],
         [4.5453e-06, 1.0000e+00],
         [6.8261e-01, 3.1739e-01],
         [1.6449e-04, 9.9984e-01],
         [9.9085e-01, 9.1517e-03]]], 4)

	y_hat = hmm.predict_proba(X, priors=priors)
	assert_array_almost_equal(y_hat, [
		[[9.6881e-02, 9.0312e-01],
         [3.8023e-03, 9.9620e-01],
         [1.1210e-08, 1.0000e+00],
         [0.0000e+00, 1.0000e+00],
         [7.5888e-01, 2.4112e-01],
         [1.2337e-15, 1.0000e+00],
         [7.9601e-01, 2.0399e-01],
         [1.1122e-06, 1.0000e+00],
         [0.0000e+00, 1.0000e+00],
         [1.0000e+00, 0.0000e+00],
         [9.9846e-01, 1.5372e-03]]], 4)


def test_hmm_predict_proba_raises(hmm, X, priors):
	X = numpy.array(X, dtype='float32')
	priors = numpy.array(priors, dtype='float32')[None, :, :]
	func = hmm.predict_proba

	assert_raises(ValueError, func, X, priors+1)
	assert_raises(ValueError, func, X, priors-1)
	assert_raises(ValueError, func, X, priors/2.0)
	assert_raises(ValueError, func, X, priors[:, :5])
	assert_raises(ValueError, func, X[:, :5], priors)
	assert_raises(ValueError, func, X, priors[:, :, 0])
	assert_raises(ValueError, func, X, priors[:, :, :1])


def test_hmm_predict_log_proba(hmm, X, priors):
	X = numpy.array([X])
	priors = priors.unsqueeze(0)

	y_hat = hmm.predict_log_proba(X)
	assert_array_almost_equal(y_hat, [
		[[-2.3343e+00, -1.0190e-01],
         [-5.5721e+00, -3.8109e-03],
         [-1.8306e+01,  0.0000e+00],
         [-1.9578e+01,  0.0000e+00],
         [-2.7591e-01, -1.4225e+00],
         [-3.3558e+01,  0.0000e+00],
         [-3.2527e-01, -1.2813e+00],
         [-1.2301e+01, -3.8147e-06],
         [-3.8184e-01, -1.1476e+00],
         [-8.7126e+00, -1.6403e-04],
         [-9.1934e-03, -4.6938e+00]]], 3)

	y_hat = hmm.predict_log_proba(X, priors=priors)
	assert_array_almost_equal(y_hat, [
		[[-2.3343e+00, -1.0190e-01],
         [-5.5721e+00, -3.8109e-03],
         [-1.8306e+01,  0.0000e+00],
         [       -inf,  0.0000e+00],
         [-2.7591e-01, -1.4225e+00],
         [-3.4329e+01,  0.0000e+00],
         [-2.2815e-01, -1.5897e+00],
         [-1.3709e+01,  0.0000e+00],
         [       -inf,  0.0000e+00],
         [ 0.0000e+00,        -inf],
         [-1.5373e-03, -6.4778e+00]]], 3)


def test_hmm_predict_log_proba_raises(hmm, X, priors):
	X = numpy.array(X, dtype='float32')
	priors = numpy.array(priors, dtype='float32')[None, :, :]
	func = hmm.predict_log_proba

	assert_raises(ValueError, func, X, priors+1)
	assert_raises(ValueError, func, X, priors-1)
	assert_raises(ValueError, func, X, priors/2.0)
	assert_raises(ValueError, func, X, priors[:, :5])
	assert_raises(ValueError, func, X[:, :5], priors)
	assert_raises(ValueError, func, X, priors[:, :, 0])
	assert_raises(ValueError, func, X, priors[:, :, :1])


def test_hmm_summarize(hmm, X, priors):
	X = numpy.array([X])
	priors = priors.unsqueeze(0)

	y_hat = hmm.summarize(X)
	assert_array_almost_equal(hmm._xw_starts_sum, [0.0969, 0.9031], 4)
	assert_array_almost_equal(hmm._xw_ends_sum, [0.9908, 0.0092], 4)
	assert_array_almost_equal(hmm._xw_sum, [[1.8028e-03, 2.2629e+00],
		[3.1568e+00, 4.5785e+00]], 4)


	y_hat = hmm.summarize(X, priors=priors)
	assert_array_almost_equal(hmm._xw_starts_sum, [0.1938, 1.8062], 4)
	assert_array_almost_equal(hmm._xw_ends_sum, [1.9893, 0.0107], 4)
	assert_array_almost_equal(hmm._xw_sum, [[1.0017, 3.9185],
		[5.714 , 9.3657]], 4)


def test_hmm_summarize_raises(hmm, X, priors):
	_test_hmm_raises(hmm, X, priors)


def test_hmm_fit_raises(hmm, X, priors):
	_test_hmm_raises(hmm, X, priors)



# independent_components.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import torch

from .._utils import _cast_as_tensor
from .._utils import _cast_as_parameter
from .._utils import _update_parameter
from .._utils import _check_parameter
from .._utils import _reshape_weights

from ._distribution import Distribution


class IndependentComponents(Distribution):
	"""An independent components distribution object.

	A distribution made up of independent, univariate, distributions that each
	model a single feature in the data. This means that instead of using a
	single type of distribution to model all of the features in your data, you
	use one distribution per feature. Note that this will likely be slower
	than using a single distribution because the amount of batching possible
	will go down significantly.

	There are two ways to initialize this object. The first is to pass in a
	set of distributions that are all initialized with parameters, at which
	point this distribution can be immediately used for inference. The second
	is to pass in a set of distributions that are not initialized with
	parameters, and then call either `fit` or `summary` + `from_summaries` to
	learn the parameters of all the distributions.


	Parameters
	----------
	distributions: list, tuple, numpy.ndarray, torch.Tensor, shape=(d,)
		An ordered iterable containing all of the distributions, one per
		feature, that will be used.

	check_data: bool, optional
		Whether to check properties of the data and potentially recast it to
		torch.tensors. This does not prevent checking of parameters but can
		slightly speed up computation when you know that your inputs are valid.
		Setting this to False is also necessary for compiling.
	"""

	def __init__(self, distributions, check_data=False):
		super().__init__(inertia=0.0, frozen=False, check_data=check_data)
		self.name = "IndependentComponents"

		if len(distributions) <= 1:
			raise ValueError("Must pass in at least 2 distributions.")
		for distribution in distributions:
			if not isinstance(distribution, Distribution):
				raise ValueError("All passed in distributions must " +
					"inherit from the Distribution object.")

		self.distributions = distributions
		self._initialized = all(d._initialized for d in distributions)
		self.d = len(distributions)
		self._reset_cache()


	def _initialize(self, d):
		"""Initialize the probability distribution.

		This method is meant to only be called internally. It initializes the
		parameters of the distribution and stores its dimensionality. For more
		complex methods, this function will do more.


		Parameters
		----------
		d: int
			The dimensionality the distribution is being initialized to.
		"""

		for distribution in self.distributions:
			distribution._initialize(d)

		self._initialized = True


	def _reset_cache(self):
		"""Reset the internally stored statistics.

		This method is meant to only be called internally. It resets the
		stored statistics used to update the model parameters as well as
		recalculates the cached values meant to speed up log probability
		calculations.
		"""

		if self._initialized == False:
			return

		for distribution in self.distributions:
			distribution._reset_cache()


	def sample(self, n):
		"""Sample from the probability distribution.

		This method will return `n` samples generated from the underlying
		probability distribution.


		Parameters
		----------
		n: int
			The number of samples to generate.
		

		Returns
		-------
		X: torch.tensor, shape=(n, self.d)
			Randomly generated samples.
		"""

		return torch.hstack([d.sample(n) for d in self.distributions])


	def log_probability(self, X):
		"""Calculate the log probability of each example.

		This method calculates the log probability of each example given the
		parameters of the distribution. The examples must be given in a 2D
		format.

		Note: This differs from some other log probability calculation
		functions, like those in torch.distributions, because it is not
		returning the log probability of each feature independently, but rather
		the total log probability of the entire example.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate.


		Returns
		-------
		logp: torch.Tensor, shape=(-1,)
			The log probability of each example.
		"""

		X = _check_parameter(_cast_as_tensor(X), "X", ndim=2, 
			shape=(-1, self.d))

		logp = torch.zeros(X.shape[0])
		for i, d in enumerate(self.distributions):
			if isinstance(X, torch.masked.MaskedTensor):
				logp.add_(d.log_probability(X[:, i:i+1])._masked_data)
			else:
				logp.add_(d.log_probability(X[:, i:i+1]))

		return logp


	def summarize(self, X, sample_weight=None):
		"""Extract the sufficient statistics from a batch of data.

		This method calculates the sufficient statistics from optionally
		weighted data and adds them to the stored cache. The examples must be
		given in a 2D format. Sample weights can either be provided as one
		value per example or as a 2D matrix of weights for each feature in
		each example.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to summarize.

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.
		"""

		if self.frozen == True:
			return
			
		X = _check_parameter(_cast_as_tensor(X), "X", ndim=2, 
			shape=(-1, self.d))

		sample_weight = _reshape_weights(X, _cast_as_tensor(sample_weight, 
			dtype=torch.float32), device=self.device)

		for i, d in enumerate(self.distributions):
			d.summarize(X[:, i:i+1], sample_weight=sample_weight[:, i:i+1])


	def from_summaries(self):
		"""Update the model parameters given the extracted statistics.

		This method uses calculated statistics from calls to the `summarize`
		method to update the distribution parameters. Hyperparameters for the
		update are passed in at initialization time.

		Note: Internally, a call to `fit` is just a successive call to the
		`summarize` method followed by the `from_summaries` method.
		"""
		
		if self.frozen == True:
			return

		for distribution in self.distributions:
			distribution.from_summaries()
